from __future__ import annotations

import os
import sys
import textwrap
import unittest
from unittest import mock

from mkdocs.config.defaults import MkDocsConfig
from mkdocs.structure.files import File, Files
from mkdocs.structure.pages import Page, _RelativePathTreeprocessor
from mkdocs.tests.base import dedent, tempdir

DOCS_DIR = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), '..', 'integration', 'subpages', 'docs'
)


def load_config(**cfg) -> MkDocsConfig:
    cfg.setdefault('site_name', 'Example')
    cfg.setdefault(
        'docs_dir',
        os.path.join(
            os.path.abspath(os.path.dirname(__file__)), '..', 'integration', 'minimal', 'docs'
        ),
    )
    conf = MkDocsConfig()
    conf.load_dict(cfg)
    errors_warnings = conf.validate()
    assert errors_warnings == ([], []), errors_warnings
    return conf


class PageTests(unittest.TestCase):
    def test_homepage(self):
        cfg = load_config(docs_dir=DOCS_DIR)
        fl = File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        self.assertIsNone(fl.page)
        pg = Page('Foo', fl, cfg)
        self.assertEqual(fl.page, pg)
        self.assertEqual(pg.url, '')
        self.assertEqual(pg.abs_url, None)
        self.assertEqual(pg.canonical_url, None)
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertTrue(pg.is_homepage)
        self.assertTrue(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertTrue(pg.is_top_level)
        self.assertEqual(pg.markdown, None)
        self.assertEqual(pg.meta, {})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, None)
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'Foo')
        self.assertEqual(pg.toc, [])

    def test_nested_index_page(self):
        cfg = load_config(docs_dir=DOCS_DIR)
        fl = File('sub1/index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page('Foo', fl, cfg)
        pg.parent = 'foo'
        self.assertEqual(pg.url, 'sub1/')
        self.assertEqual(pg.abs_url, None)
        self.assertEqual(pg.canonical_url, None)
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertFalse(pg.is_homepage)
        self.assertTrue(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertFalse(pg.is_top_level)
        self.assertEqual(pg.markdown, None)
        self.assertEqual(pg.meta, {})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, 'foo')
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'Foo')
        self.assertEqual(pg.toc, [])

    def test_nested_index_page_no_parent(self):
        cfg = load_config(docs_dir=DOCS_DIR)
        fl = File('sub1/index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page('Foo', fl, cfg)
        pg.parent = None  # non-homepage at nav root level; see #1919.
        self.assertEqual(pg.url, 'sub1/')
        self.assertEqual(pg.abs_url, None)
        self.assertEqual(pg.canonical_url, None)
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertFalse(pg.is_homepage)
        self.assertTrue(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertTrue(pg.is_top_level)
        self.assertEqual(pg.markdown, None)
        self.assertEqual(pg.meta, {})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, None)
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'Foo')
        self.assertEqual(pg.toc, [])

    def test_nested_index_page_no_parent_no_directory_urls(self):
        cfg = load_config(docs_dir=DOCS_DIR, use_directory_urls=False)
        fl = File('sub1/index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page('Foo', fl, cfg)
        pg.parent = None  # non-homepage at nav root level; see #1919.
        self.assertEqual(pg.url, 'sub1/index.html')
        self.assertEqual(pg.abs_url, None)
        self.assertEqual(pg.canonical_url, None)
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertFalse(pg.is_homepage)
        self.assertTrue(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertTrue(pg.is_top_level)
        self.assertEqual(pg.markdown, None)
        self.assertEqual(pg.meta, {})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, None)
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'Foo')
        self.assertEqual(pg.toc, [])

    def test_nested_nonindex_page(self):
        cfg = load_config(docs_dir=DOCS_DIR)
        fl = File('sub1/non-index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page('Foo', fl, cfg)
        pg.parent = 'foo'
        self.assertEqual(pg.url, 'sub1/non-index/')
        self.assertEqual(pg.abs_url, None)
        self.assertEqual(pg.canonical_url, None)
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertFalse(pg.is_homepage)
        self.assertFalse(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertFalse(pg.is_top_level)
        self.assertEqual(pg.markdown, None)
        self.assertEqual(pg.meta, {})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, 'foo')
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'Foo')
        self.assertEqual(pg.toc, [])

    def test_page_defaults(self):
        cfg = load_config()
        fl = File('testing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page('Foo', fl, cfg)
        self.assertRegex(pg.update_date, r'\d{4}-\d{2}-\d{2}')
        self.assertEqual(pg.url, 'testing/')
        self.assertEqual(pg.abs_url, None)
        self.assertEqual(pg.canonical_url, None)
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertFalse(pg.is_homepage)
        self.assertFalse(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertTrue(pg.is_top_level)
        self.assertEqual(pg.markdown, None)
        self.assertEqual(pg.meta, {})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, None)
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'Foo')
        self.assertEqual(pg.toc, [])

    def test_page_no_directory_url(self):
        cfg = load_config(use_directory_urls=False)
        fl = File('testing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page('Foo', fl, cfg)
        self.assertEqual(pg.url, 'testing.html')
        self.assertEqual(pg.abs_url, None)
        self.assertEqual(pg.canonical_url, None)
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertFalse(pg.is_homepage)
        self.assertFalse(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertTrue(pg.is_top_level)
        self.assertEqual(pg.markdown, None)
        self.assertEqual(pg.meta, {})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, None)
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'Foo')
        self.assertEqual(pg.toc, [])

    def test_page_canonical_url(self):
        cfg = load_config(site_url='http://example.com')
        fl = File('testing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page('Foo', fl, cfg)
        self.assertEqual(pg.url, 'testing/')
        self.assertEqual(pg.abs_url, '/testing/')
        self.assertEqual(pg.canonical_url, 'http://example.com/testing/')
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertFalse(pg.is_homepage)
        self.assertFalse(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertTrue(pg.is_top_level)
        self.assertEqual(pg.markdown, None)
        self.assertEqual(pg.meta, {})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, None)
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'Foo')
        self.assertEqual(pg.toc, [])

    def test_page_canonical_url_nested(self):
        cfg = load_config(site_url='http://example.com/foo/')
        fl = File('testing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page('Foo', fl, cfg)
        self.assertEqual(pg.url, 'testing/')
        self.assertEqual(pg.abs_url, '/foo/testing/')
        self.assertEqual(pg.canonical_url, 'http://example.com/foo/testing/')
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertFalse(pg.is_homepage)
        self.assertFalse(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertTrue(pg.is_top_level)
        self.assertEqual(pg.markdown, None)
        self.assertEqual(pg.meta, {})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, None)
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'Foo')
        self.assertEqual(pg.toc, [])

    def test_page_canonical_url_nested_no_slash(self):
        cfg = load_config(site_url='http://example.com/foo')
        fl = File('testing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page('Foo', fl, cfg)
        self.assertEqual(pg.url, 'testing/')
        self.assertEqual(pg.abs_url, '/foo/testing/')
        self.assertEqual(pg.canonical_url, 'http://example.com/foo/testing/')
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertFalse(pg.is_homepage)
        self.assertFalse(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertTrue(pg.is_top_level)
        self.assertEqual(pg.markdown, None)
        self.assertEqual(pg.meta, {})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, None)
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'Foo')
        self.assertEqual(pg.toc, [])

    def test_predefined_page_title(self):
        cfg = load_config()
        fl = File('testing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page('Page Title', fl, cfg)
        pg.read_source(cfg)
        self.assertEqual(pg.url, 'testing/')
        self.assertEqual(pg.abs_url, None)
        self.assertEqual(pg.canonical_url, None)
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertFalse(pg.is_homepage)
        self.assertFalse(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertTrue(pg.is_top_level)
        self.assertTrue(pg.markdown.startswith('# Welcome to MkDocs\n'))
        self.assertEqual(pg.meta, {})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, None)
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'Page Title')
        self.assertEqual(pg.toc, [])

    def test_page_title_from_markdown(self):
        cfg = load_config()
        fl = File('testing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page(None, fl, cfg)
        pg.read_source(cfg)
        self.assertEqual(pg.url, 'testing/')
        self.assertEqual(pg.abs_url, None)
        self.assertEqual(pg.canonical_url, None)
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertFalse(pg.is_homepage)
        self.assertFalse(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertTrue(pg.is_top_level)
        self.assertTrue(pg.markdown.startswith('# Welcome to MkDocs\n'))
        self.assertEqual(pg.meta, {})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, None)
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'Welcome to MkDocs')
        pg.render(cfg, fl)
        self.assertEqual(pg.title, 'Welcome to MkDocs')

    _SETEXT_CONTENT = dedent(
        '''
        Welcome to MkDocs Setext
        ========================

        This tests extracting a setext style title.
        '''
    )

    @tempdir(files={'testing_setext_title.md': _SETEXT_CONTENT})
    def test_page_title_from_setext_markdown(self, docs_dir):
        cfg = load_config()
        fl = File('testing_setext_title.md', docs_dir, docs_dir, use_directory_urls=True)
        pg = Page(None, fl, cfg)
        self.assertIsNone(pg.title)
        pg.read_source(cfg)
        self.assertEqual(pg.title, 'Testing setext title')
        pg.render(cfg, fl)
        self.assertEqual(pg.title, 'Welcome to MkDocs Setext')

    @tempdir(files={'testing_setext_title.md': _SETEXT_CONTENT})
    def test_page_title_from_markdown_stripped_anchorlinks(self, docs_dir):
        cfg = MkDocsConfig()
        cfg.site_name = 'example'
        cfg.markdown_extensions = {'toc': {'permalink': '&'}}
        self.assertEqual(cfg.validate(), ([], []))
        fl = File('testing_setext_title.md', docs_dir, docs_dir, use_directory_urls=True)
        pg = Page(None, fl, cfg)
        pg.read_source(cfg)
        pg.render(cfg, fl)
        self.assertEqual(pg.title, 'Welcome to MkDocs Setext')

    _FORMATTING_CONTENT = dedent(
        '''
        # \\*Hello --- *beautiful* `world`

        Hi.
        '''
    )

    @tempdir(files={'testing_formatting.md': _FORMATTING_CONTENT})
    def test_page_title_from_markdown_strip_formatting(self, docs_dir):
        cfg = load_config()
        cfg.markdown_extensions.append('smarty')
        fl = File('testing_formatting.md', docs_dir, docs_dir, use_directory_urls=True)
        pg = Page(None, fl, cfg)
        pg.read_source(cfg)
        pg.render(cfg, fl)
        self.assertEqual(pg.title, '*Hello &mdash; beautiful world')

    _ATTRLIST_CONTENT = dedent(
        '''
        # Welcome to MkDocs Attr { #welcome }

        This tests extracting the title, with enabled attr_list markdown_extension.
        '''
    )

    @tempdir(files={'testing_attr_list.md': _ATTRLIST_CONTENT})
    def test_page_title_from_markdown_stripped_attr_list(self, docs_dir):
        cfg = load_config()
        cfg.markdown_extensions.append('attr_list')
        fl = File('testing_attr_list.md', docs_dir, docs_dir, use_directory_urls=True)
        pg = Page(None, fl, cfg)
        pg.read_source(cfg)
        pg.render(cfg, fl)
        self.assertEqual(pg.title, 'Welcome to MkDocs Attr')

    @tempdir(files={'testing_attr_list.md': _ATTRLIST_CONTENT})
    def test_page_title_from_markdown_preserved_attr_list(self, docs_dir):
        cfg = load_config()
        fl = File('testing_attr_list.md', docs_dir, docs_dir, use_directory_urls=True)
        pg = Page(None, fl, cfg)
        pg.read_source(cfg)
        pg.render(cfg, fl)
        self.assertEqual(pg.title, 'Welcome to MkDocs Attr { #welcome }')

    def test_page_title_from_meta(self):
        cfg = load_config(docs_dir=DOCS_DIR)
        fl = File('metadata.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page(None, fl, cfg)
        pg.read_source(cfg)
        self.assertEqual(pg.url, 'metadata/')
        self.assertEqual(pg.abs_url, None)
        self.assertEqual(pg.canonical_url, None)
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertFalse(pg.is_homepage)
        self.assertFalse(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertTrue(pg.is_top_level)
        self.assertTrue(pg.markdown.startswith('# Welcome to MkDocs\n'))
        self.assertEqual(pg.meta, {'title': 'A Page Title'})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, None)
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'A Page Title')
        self.assertEqual(pg.toc, [])
        pg.render(cfg, fl)
        self.assertEqual(pg.title, 'A Page Title')

    def test_page_title_from_filename(self):
        cfg = load_config(docs_dir=DOCS_DIR)
        fl = File('page-title.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page(None, fl, cfg)
        pg.read_source(cfg)
        self.assertEqual(pg.url, 'page-title/')
        self.assertEqual(pg.abs_url, None)
        self.assertEqual(pg.canonical_url, None)
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertFalse(pg.is_homepage)
        self.assertFalse(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertTrue(pg.is_top_level)
        self.assertTrue(pg.markdown.startswith('Page content.\n'))
        self.assertEqual(pg.meta, {})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, None)
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'Page title')
        pg.render(cfg, fl)
        self.assertEqual(pg.title, 'Page title')

    def test_page_title_from_capitalized_filename(self):
        cfg = load_config(docs_dir=DOCS_DIR)
        fl = File('pageTitle.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page(None, fl, cfg)
        pg.read_source(cfg)
        self.assertEqual(pg.url, 'pageTitle/')
        self.assertEqual(pg.abs_url, None)
        self.assertEqual(pg.canonical_url, None)
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertFalse(pg.is_homepage)
        self.assertFalse(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertTrue(pg.is_top_level)
        self.assertTrue(pg.markdown.startswith('Page content.\n'))
        self.assertEqual(pg.meta, {})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, None)
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'pageTitle')

    def test_page_title_from_homepage_filename(self):
        cfg = load_config(docs_dir=DOCS_DIR)
        fl = File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page(None, fl, cfg)
        pg.read_source(cfg)
        self.assertEqual(pg.url, '')
        self.assertEqual(pg.abs_url, None)
        self.assertEqual(pg.canonical_url, None)
        self.assertEqual(pg.edit_url, None)
        self.assertEqual(pg.file, fl)
        self.assertEqual(pg.content, None)
        self.assertTrue(pg.is_homepage)
        self.assertTrue(pg.is_index)
        self.assertTrue(pg.is_page)
        self.assertFalse(pg.is_section)
        self.assertTrue(pg.is_top_level)
        self.assertTrue(pg.markdown.startswith('## Test'))
        self.assertEqual(pg.meta, {})
        self.assertEqual(pg.next_page, None)
        self.assertEqual(pg.parent, None)
        self.assertEqual(pg.previous_page, None)
        self.assertEqual(pg.title, 'Home')
        self.assertEqual(pg.toc, [])

    def test_page_eq(self):
        cfg = load_config()
        fl = File('testing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page('Foo', fl, cfg)
        self.assertTrue(pg == Page('Foo', fl, cfg))

    def test_page_ne(self):
        cfg = load_config()
        f1 = File('testing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        f2 = File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page('Foo', f1, cfg)
        # Different Title
        self.assertTrue(pg != Page('Bar', f1, cfg))
        # Different File
        self.assertTrue(pg != Page('Foo', f2, cfg))

    @tempdir()
    def test_BOM(self, docs_dir):
        md_src = '# An UTF-8 encoded file with a BOM'
        cfg = load_config(docs_dir=docs_dir)
        fl = File('index.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page(None, fl, cfg)
        # Create an UTF-8 Encoded file with BOM (as Microsoft editors do). See #1186
        with open(fl.abs_src_path, 'w', encoding='utf-8-sig') as f:
            f.write(md_src)
        # Now read the file.
        pg.read_source(cfg)
        # Ensure the BOM (`\ufeff`) is removed
        self.assertNotIn('\ufeff', pg.markdown)
        self.assertEqual(pg.markdown, md_src)
        self.assertEqual(pg.meta, {})

    def test_page_edit_url(
        self, paths={'testing.md': 'testing/', 'sub1/non-index.md': 'sub1/non-index/'}
    ):
        for case in [
            dict(
                config={'repo_url': 'http://github.com/mkdocs/mkdocs'},
                edit_url='http://github.com/mkdocs/mkdocs/edit/master/docs/testing.md',
                edit_url2='http://github.com/mkdocs/mkdocs/edit/master/docs/sub1/non-index.md',
            ),
            dict(
                config={'repo_url': 'https://github.com/mkdocs/mkdocs/'},
                edit_url='https://github.com/mkdocs/mkdocs/edit/master/docs/testing.md',
                edit_url2='https://github.com/mkdocs/mkdocs/edit/master/docs/sub1/non-index.md',
            ),
            dict(
                config={'repo_url': 'http://example.com'},
                edit_url=None,
                edit_url2=None,
            ),
            dict(
                config={'repo_url': 'http://example.com', 'edit_uri': 'edit/master'},
                edit_url='http://example.com/edit/master/testing.md',
                edit_url2='http://example.com/edit/master/sub1/non-index.md',
            ),
            dict(
                config={'repo_url': 'http://example.com', 'edit_uri': '/edit/master'},
                edit_url='http://example.com/edit/master/testing.md',
                edit_url2='http://example.com/edit/master/sub1/non-index.md',
            ),
            dict(
                config={'repo_url': 'http://example.com/foo/', 'edit_uri': '/edit/master/'},
                edit_url='http://example.com/edit/master/testing.md',
                edit_url2='http://example.com/edit/master/sub1/non-index.md',
            ),
            dict(
                config={'repo_url': 'http://example.com/foo', 'edit_uri': '/edit/master/'},
                edit_url='http://example.com/edit/master/testing.md',
                edit_url2='http://example.com/edit/master/sub1/non-index.md',
            ),
            dict(
                config={'repo_url': 'http://example.com/foo/', 'edit_uri': '/edit/master'},
                edit_url='http://example.com/edit/master/testing.md',
                edit_url2='http://example.com/edit/master/sub1/non-index.md',
            ),
            dict(
                config={'repo_url': 'http://example.com/foo/', 'edit_uri': 'edit/master/'},
                edit_url='http://example.com/foo/edit/master/testing.md',
                edit_url2='http://example.com/foo/edit/master/sub1/non-index.md',
            ),
            dict(
                config={'repo_url': 'http://example.com/foo', 'edit_uri': 'edit/master/'},
                edit_url='http://example.com/foo/edit/master/testing.md',
                edit_url2='http://example.com/foo/edit/master/sub1/non-index.md',
            ),
            dict(
                config={'repo_url': 'http://example.com', 'edit_uri': '?query=edit/master'},
                edit_url='http://example.com?query=edit/master/testing.md',
                edit_url2='http://example.com?query=edit/master/sub1/non-index.md',
            ),
            dict(
                config={'repo_url': 'http://example.com/', 'edit_uri': '?query=edit/master/'},
                edit_url='http://example.com/?query=edit/master/testing.md',
                edit_url2='http://example.com/?query=edit/master/sub1/non-index.md',
            ),
            dict(
                config={'repo_url': 'http://example.com', 'edit_uri': '#edit/master'},
                edit_url='http://example.com#edit/master/testing.md',
                edit_url2='http://example.com#edit/master/sub1/non-index.md',
            ),
            dict(
                config={'repo_url': 'http://example.com/', 'edit_uri': '#edit/master/'},
                edit_url='http://example.com/#edit/master/testing.md',
                edit_url2='http://example.com/#edit/master/sub1/non-index.md',
            ),
            dict(
                config={'edit_uri': 'http://example.com/edit/master'},
                edit_url='http://example.com/edit/master/testing.md',
                edit_url2='http://example.com/edit/master/sub1/non-index.md',
            ),
            dict(
                config={'edit_uri_template': 'https://github.com/project/repo/wiki/{path_noext}'},
                edit_url='https://github.com/project/repo/wiki/testing',
                edit_url2='https://github.com/project/repo/wiki/sub1/non-index',
            ),
            dict(
                config={
                    'repo_url': 'https://github.com/project/repo/wiki',
                    'edit_uri_template': '{path_noext}/_edit',
                },
                edit_url='https://github.com/project/repo/wiki/testing/_edit',
                edit_url2='https://github.com/project/repo/wiki/sub1/non-index/_edit',
            ),
            dict(
                config={
                    'repo_url': 'https://gitlab.com/project/repo',
                    'edit_uri_template': '-/sse/master/docs%2F{path!q}',
                },
                edit_url='https://gitlab.com/project/repo/-/sse/master/docs%2Ftesting.md',
                edit_url2='https://gitlab.com/project/repo/-/sse/master/docs%2Fsub1%2Fnon-index.md',
            ),
            dict(
                config={
                    'repo_url': 'https://bitbucket.org/project/repo/',
                    'edit_uri_template': 'src/master/docs/{path}?mode=edit',
                },
                edit_url='https://bitbucket.org/project/repo/src/master/docs/testing.md?mode=edit',
                edit_url2='https://bitbucket.org/project/repo/src/master/docs/sub1/non-index.md?mode=edit',
            ),
            dict(
                config={
                    'repo_url': 'http://example.com',
                    'edit_uri': '',
                    'edit_uri_template': '',
                },  # Set to blank value
                edit_url=None,
                edit_url2=None,
            ),
            dict(config={}, edit_url=None, edit_url2=None),  # Nothing defined
        ]:
            for i, path in enumerate(paths, 1):
                edit_url_key = f'edit_url{i}' if i > 1 else 'edit_url'
                with self.subTest(case['config'], path=path):
                    cfg = load_config(**case['config'])
                    fl = File(path, cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
                    pg = Page('Foo', fl, cfg)
                    self.assertEqual(pg.url, paths[path])
                    self.assertEqual(pg.edit_url, case[edit_url_key])

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def test_page_edit_url_windows(self):
        self.test_page_edit_url(
            paths={'testing.md': 'testing/', 'sub1\\non-index.md': 'sub1/non-index/'}
        )

    def test_page_edit_url_warning(self):
        for case in [
            dict(
                config={'edit_uri': 'edit/master'},
                edit_url='edit/master/testing.md',
                warning="WARNING:mkdocs.structure.pages:edit_uri: "
                "'edit/master/testing.md' is not a valid URL, it should include the http:// (scheme)",
            ),
        ]:
            with self.subTest(case['config']):
                with self.assertLogs('mkdocs') as cm:
                    cfg = load_config(**case['config'])
                    fl = File('testing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
                    pg = Page('Foo', fl, cfg)
                self.assertEqual(pg.url, 'testing/')
                self.assertEqual(pg.edit_url, case['edit_url'])
                self.assertEqual(cm.output, [case['warning']])

    def test_page_render(self):
        cfg = load_config()
        fl = File('testing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page('Foo', fl, cfg)
        pg.read_source(cfg)
        self.assertEqual(pg.content, None)
        self.assertEqual(pg.toc, [])
        pg.render(cfg, [fl])
        self.assertTrue(
            pg.content.startswith('<h1 id="welcome-to-mkdocs">Welcome to MkDocs</h1>\n')
        )
        self.assertEqual(
            str(pg.toc).strip(),
            dedent(
                """
                Welcome to MkDocs - #welcome-to-mkdocs
                    Commands - #commands
                    Project layout - #project-layout
                """
            ),
        )

    def test_missing_page(self):
        cfg = load_config()
        fl = File('missing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page('Foo', fl, cfg)
        with self.assertLogs('mkdocs') as cm:
            with self.assertRaises(OSError):
                pg.read_source(cfg)
        self.assertEqual(
            '\n'.join(cm.output), 'ERROR:mkdocs.structure.pages:File not found: missing.md'
        )


class SourceDateEpochTests(unittest.TestCase):
    def setUp(self):
        self.default = os.environ.get('SOURCE_DATE_EPOCH', None)
        os.environ['SOURCE_DATE_EPOCH'] = '0'

    def test_source_date_epoch(self):
        cfg = load_config()
        fl = File('testing.md', cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
        pg = Page('Foo', fl, cfg)
        self.assertEqual(pg.update_date, '1970-01-01')

    def tearDown(self):
        if self.default is not None:
            os.environ['SOURCE_DATE_EPOCH'] = self.default
        else:
            del os.environ['SOURCE_DATE_EPOCH']


class RelativePathExtensionTests(unittest.TestCase):
    def get_rendered_result(
        self, *, content: str, files: list[str], logs: str = '', **kwargs
    ) -> str:
        cfg = load_config(docs_dir=DOCS_DIR, **kwargs)
        fs = [File(f, cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls) for f in files]
        pg = Page('Foo', fs[0], cfg)

        with mock.patch('mkdocs.structure.pages.open', mock.mock_open(read_data=content)):
            pg.read_source(cfg)
        if logs:
            with self.assertLogs('mkdocs.structure.pages') as cm:
                pg.render(cfg, Files(fs))
            msgs = [f'{r.levelname}:{r.message}' for r in cm.records]
            self.assertEqual('\n'.join(msgs), textwrap.dedent(logs).strip('\n'))
        elif sys.version_info >= (3, 10):
            with self.assertNoLogs('mkdocs.structure.pages'):
                pg.render(cfg, Files(fs))
        else:
            pg.render(cfg, Files(fs))

        assert pg.content is not None
        content = pg.content
        if content.startswith('<p>') and content.endswith('</p>'):
            content = content[3:-4]
        return content

    def test_relative_html_link(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[link](non-index.md)', files=['index.md', 'non-index.md']
            ),
            '<a href="non-index/">link</a>',
        )
        self.assertEqual(
            self.get_rendered_result(
                use_directory_urls=False,
                content='[link](non-index.md)',
                files=['index.md', 'non-index.md'],
            ),
            '<a href="non-index.html">link</a>',
        )

    def test_relative_html_link_index(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[link](index.md)', files=['non-index.md', 'index.md']
            ),
            '<a href="../">link</a>',
        )
        self.assertEqual(
            self.get_rendered_result(
                use_directory_urls=False,
                content='[link](index.md)',
                files=['non-index.md', 'index.md'],
            ),
            '<a href="index.html">link</a>',
        )

    def test_relative_html_link_sub_index(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[link](sub2/index.md)', files=['index.md', 'sub2/index.md']
            ),
            '<a href="sub2/">link</a>',
        )
        self.assertEqual(
            self.get_rendered_result(
                use_directory_urls=False,
                content='[link](sub2/index.md)',
                files=['index.md', 'sub2/index.md'],
            ),
            '<a href="sub2/index.html">link</a>',
        )

    def test_relative_html_link_sub_page(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[link](sub2/non-index.md)', files=['index.md', 'sub2/non-index.md']
            ),
            '<a href="sub2/non-index/">link</a>',
        )
        self.assertEqual(
            self.get_rendered_result(
                use_directory_urls=False,
                content='[link](sub2/non-index.md)',
                files=['index.md', 'sub2/non-index.md'],
            ),
            '<a href="sub2/non-index.html">link</a>',
        )

    def test_relative_doc_link_without_extension(self):
        self.assertEqual(
            self.get_rendered_result(
                use_directory_urls=False,
                content='[link](bar/Dockerfile)',
                files=['foo/bar.md', 'foo/bar/Dockerfile'],
            ),
            '<a href="bar/Dockerfile">link</a>',
        )
        self.assertEqual(
            self.get_rendered_result(
                content='[link](bar/Dockerfile)',
                files=['foo/bar.md', 'foo/bar/Dockerfile'],
            ),
            '<a href="Dockerfile">link</a>',
        )

    def test_relative_html_link_with_encoded_space(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[link](file%20name.md)', files=['index.md', 'file name.md']
            ),
            '<a href="file%20name/">link</a>',
        )

    def test_relative_html_link_with_unencoded_space(self):
        self.assertEqual(
            self.get_rendered_result(
                use_directory_urls=False,
                content='[link](file name.md)',
                files=['index.md', 'file name.md'],
            ),
            '<a href="file%20name.html">link</a>',
        )

    def test_relative_html_link_parent_index(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[link](../index.md)', files=['sub2/non-index.md', 'index.md']
            ),
            '<a href="../../">link</a>',
        )
        self.assertEqual(
            self.get_rendered_result(
                use_directory_urls=False,
                content='[link](../index.md)',
                files=['sub2/non-index.md', 'index.md'],
            ),
            '<a href="../index.html">link</a>',
        )

    def test_relative_html_link_hash(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[link](non-index.md#hash)', files=['index.md', 'non-index.md']
            ),
            '<a href="non-index/#hash">link</a>',
        )

    def test_relative_html_link_sub_index_hash(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[link](sub2/index.md#hash)', files=['index.md', 'sub2/index.md']
            ),
            '<a href="sub2/#hash">link</a>',
        )
        self.assertEqual(
            self.get_rendered_result(
                use_directory_urls=False,
                content='[link](sub2/index.md#hash)',
                files=['index.md', 'sub2/index.md'],
            ),
            '<a href="sub2/index.html#hash">link</a>',
        )

    def test_relative_html_link_sub_page_hash(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[link](sub2/non-index.md#hash)', files=['index.md', 'sub2/non-index.md']
            ),
            '<a href="sub2/non-index/#hash">link</a>',
        )

    def test_relative_html_link_hash_only(self):
        for use_directory_urls in True, False:
            self.assertEqual(
                self.get_rendered_result(
                    use_directory_urls=use_directory_urls,
                    content='[link](#hash)',
                    files=['index.md'],
                ),
                '<a href="#hash">link</a>',
            )

    def test_relative_image_link_from_homepage(self):
        for use_directory_urls in True, False:
            self.assertEqual(
                self.get_rendered_result(
                    use_directory_urls=use_directory_urls,
                    content='![image](image.png)',
                    files=['index.md', 'image.png'],
                ),
                '<img alt="image" src="image.png" />',  # no opening ./
            )

    def test_relative_image_link_from_subpage(self):
        self.assertEqual(
            self.get_rendered_result(
                content='![image](../image.png)', files=['sub2/non-index.md', 'image.png']
            ),
            '<img alt="image" src="../../image.png" />',
        )

    def test_relative_image_link_from_sibling(self):
        self.assertEqual(
            self.get_rendered_result(
                content='![image](image.png)', files=['non-index.md', 'image.png']
            ),
            '<img alt="image" src="../image.png" />',
        )
        self.assertEqual(
            self.get_rendered_result(
                use_directory_urls=False,
                content='![image](image.png)',
                files=['non-index.md', 'image.png'],
            ),
            '<img alt="image" src="image.png" />',
        )

    def test_no_links(self):
        self.assertEqual(
            self.get_rendered_result(content='*__not__ a link*.', files=['index.md']),
            '<em><strong>not</strong> a link</em>.',
        )

    def test_bad_relative_doc_link(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[link](non-existent.md)',
                files=['index.md'],
                logs="WARNING:Doc file 'index.md' contains a link 'non-existent.md', but the target is not found among documentation files.",
            ),
            '<a href="non-existent.md">link</a>',
        )
        self.assertEqual(
            self.get_rendered_result(
                validation=dict(links=dict(not_found='info')),
                content='[link](../non-existent.md)',
                files=['sub/index.md'],
                logs="INFO:Doc file 'sub/index.md' contains a link '../non-existent.md', but the target 'non-existent.md' is not found among documentation files.",
            ),
            '<a href="../non-existent.md">link</a>',
        )

    def test_relative_slash_link_with_suggestion(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[link](../about/)',
                files=['foo/index.md', 'about.md'],
                logs="INFO:Doc file 'foo/index.md' contains an unrecognized relative link '../about/', it was left as is. Did you mean '../about.md'?",
            ),
            '<a href="../about/">link</a>',
        )
        self.assertEqual(
            self.get_rendered_result(
                validation=dict(links=dict(unrecognized_links='warn')),
                content='[link](../#example)',
                files=['foo/bar.md', 'index.md'],
                logs="WARNING:Doc file 'foo/bar.md' contains an unrecognized relative link '../#example', it was left as is. Did you mean '../index.md#example'?",
            ),
            '<a href="../#example">link</a>',
        )

    def test_self_anchor_link_with_suggestion(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[link](./#test)',
                files=['index.md'],
                logs="INFO:Doc file 'index.md' contains an unrecognized relative link './#test', it was left as is. Did you mean '#test'?",
            ),
            '<a href="./#test">link</a>',
        )

    def test_absolute_self_anchor_link_with_suggestion(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[link](/index#test)',
                files=['index.md'],
                logs="INFO:Doc file 'index.md' contains an absolute link '/index#test', it was left as is. Did you mean '#test'?",
            ),
            '<a href="/index#test">link</a>',
        )

    def test_absolute_self_anchor_link_with_validation_and_suggestion(self):
        self.assertEqual(
            self.get_rendered_result(
                validation=dict(links=dict(absolute_links='relative_to_docs')),
                content='[link](/index#test)',
                files=['index.md'],
                logs="WARNING:Doc file 'index.md' contains a link '/index#test', but the target 'index' is not found among documentation files. Did you mean '#test'?",
            ),
            '<a href="/index#test">link</a>',
        )

    def test_absolute_anchor_link_with_validation(self):
        self.assertEqual(
            self.get_rendered_result(
                validation=dict(links=dict(absolute_links='relative_to_docs')),
                content='[link](/foo/bar.md#test)',
                files=['index.md', 'foo/bar.md'],
            ),
            '<a href="foo/bar/#test">link</a>',
        )

    def test_absolute_anchor_link_with_validation_and_suggestion(self):
        self.assertEqual(
            self.get_rendered_result(
                validation=dict(links=dict(absolute_links='relative_to_docs')),
                content='[link](/foo/bar#test)',
                files=['zoo/index.md', 'foo/bar.md'],
                logs="WARNING:Doc file 'zoo/index.md' contains a link '/foo/bar#test', but the target 'foo/bar' is not found among documentation files. Did you mean '/foo/bar.md#test'?",
            ),
            '<a href="/foo/bar#test">link</a>',
        )

    def test_external_link(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[external](http://example.com/index.md)', files=['index.md']
            ),
            '<a href="http://example.com/index.md">external</a>',
        )

    def test_absolute_link_with_suggestion(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[absolute link](/path/to/file.md)',
                files=['index.md', 'path/to/file.md'],
                logs="INFO:Doc file 'index.md' contains an absolute link '/path/to/file.md', it was left as is. Did you mean 'path/to/file.md'?",
            ),
            '<a href="/path/to/file.md">absolute link</a>',
        )
        self.assertEqual(
            self.get_rendered_result(
                use_directory_urls=False,
                content='[absolute link](/path/to/file/)',
                files=['path/index.md', 'path/to/file.md'],
                logs="INFO:Doc file 'path/index.md' contains an absolute link '/path/to/file/', it was left as is.",
            ),
            '<a href="/path/to/file/">absolute link</a>',
        )
        self.assertEqual(
            self.get_rendered_result(
                content='[absolute link](/path/to/file)',
                files=['path/index.md', 'path/to/file.md'],
                logs="INFO:Doc file 'path/index.md' contains an absolute link '/path/to/file', it was left as is. Did you mean 'to/file.md'?",
            ),
            '<a href="/path/to/file">absolute link</a>',
        )

    def test_absolute_link_with_validation(self):
        self.assertEqual(
            self.get_rendered_result(
                validation=dict(links=dict(absolute_links='relative_to_docs')),
                content='[absolute link](/path/to/file.md)',
                files=['index.md', 'path/to/file.md'],
            ),
            '<a href="path/to/file/">absolute link</a>',
        )
        self.assertEqual(
            self.get_rendered_result(
                validation=dict(links=dict(absolute_links='relative_to_docs')),
                use_directory_urls=False,
                content='[absolute link](/path/to/file.md)',
                files=['path/index.md', 'path/to/file.md'],
            ),
            '<a href="to/file.html">absolute link</a>',
        )

    def test_absolute_link_with_validation_and_suggestion(self):
        self.assertEqual(
            self.get_rendered_result(
                validation=dict(links=dict(absolute_links='relative_to_docs')),
                use_directory_urls=False,
                content='[absolute link](/path/to/file/)',
                files=['path/index.md', 'path/to/file.md'],
                logs="WARNING:Doc file 'path/index.md' contains a link '/path/to/file/', but the target 'path/to/file' is not found among documentation files.",
            ),
            '<a href="/path/to/file/">absolute link</a>',
        )
        self.assertEqual(
            self.get_rendered_result(
                validation=dict(links=dict(absolute_links='relative_to_docs')),
                content='[absolute link](/path/to/file)',
                files=['path/index.md', 'path/to/file.md'],
                logs="WARNING:Doc file 'path/index.md' contains a link '/path/to/file', but the target is not found among documentation files. Did you mean '/path/to/file.md'?",
            ),
            '<a href="/path/to/file">absolute link</a>',
        )

    def test_absolute_link_with_validation_just_slash(self):
        self.assertEqual(
            self.get_rendered_result(
                validation=dict(links=dict(absolute_links='relative_to_docs')),
                content='[absolute link](/)',
                files=['path/to/file.md', 'index.md'],
                logs="WARNING:Doc file 'path/to/file.md' contains a link '/', but the target '.' is not found among documentation files. Did you mean '/index.md'?",
            ),
            '<a href="/">absolute link</a>',
        )

    def test_absolute_link_preserved_and_warned(self):
        self.assertEqual(
            self.get_rendered_result(
                validation=dict(links=dict(absolute_links='warn')),
                content='[absolute link](/path/to/file.md)',
                files=['index.md'],
                logs="WARNING:Doc file 'index.md' contains an absolute link '/path/to/file.md', it was left as is.",
            ),
            '<a href="/path/to/file.md">absolute link</a>',
        )
        self.assertEqual(
            self.get_rendered_result(
                validation=dict(links=dict(absolute_links='ignore')),
                content='[absolute link](/path/to/file.md)',
                files=['index.md'],
            ),
            '<a href="/path/to/file.md">absolute link</a>',
        )

    def test_image_link_with_suggestion(self):
        self.assertEqual(
            self.get_rendered_result(
                content='![image](../image.png)',
                files=['foo/bar.md', 'foo/image.png'],
                logs="WARNING:Doc file 'foo/bar.md' contains a link '../image.png', but the target 'image.png' is not found among documentation files. Did you mean 'image.png'?",
            ),
            '<img alt="image" src="../image.png" />',
        )
        self.assertEqual(
            self.get_rendered_result(
                content='![image](/image.png)',
                files=['foo/bar.md', 'image.png'],
                logs="INFO:Doc file 'foo/bar.md' contains an absolute link '/image.png', it was left as is. Did you mean '../image.png'?",
            ),
            '<img alt="image" src="/image.png" />',
        )

    def test_absolute_win_local_path(self):
        for use_directory_urls in True, False:
            self.assertEqual(
                self.get_rendered_result(
                    use_directory_urls=use_directory_urls,
                    content='[absolute local path](\\image.png)',
                    files=['index.md'],
                    logs="INFO:Doc file 'index.md' contains an absolute link '\\image.png', it was left as is.",
                ),
                '<a href="\\image.png">absolute local path</a>',
            )

    def test_email_link(self):
        self.assertEqual(
            self.get_rendered_result(content='<mail@example.com>', files=['index.md']),
            # Markdown's default behavior is to obscure email addresses by entity-encoding them.
            # The following is equivalent to: '<a href="mailto:mail@example.com">mail@example.com</a>'
            '<a href="&#109;&#97;&#105;&#108;&#116;&#111;&#58;&#109;&#97;&#105;&#108;&#64;&#101;'
            '&#120;&#97;&#109;&#112;&#108;&#101;&#46;&#99;&#111;&#109;">&#109;&#97;&#105;&#108;&#64;'
            '&#101;&#120;&#97;&#109;&#112;&#108;&#101;&#46;&#99;&#111;&#109;</a>',
        )

    def test_invalid_email_link(self):
        self.assertEqual(
            self.get_rendered_result(
                content='[contact](mail@example.com)',
                files=['index.md'],
                logs="WARNING:Doc file 'index.md' contains a link 'mail@example.com', but the target is not found among documentation files. Did you mean 'mailto:mail@example.com'?",
            ),
            '<a href="mail@example.com">contact</a>',
        )

    def test_possible_target_uris(self):
        def test(paths, expected='', exp_true=None, exp_false=None):
            """Test that `possible_target_uris` yields expected values, for use_directory_urls = true and false."""
            for use_directory_urls, expected_paths in (
                (True, exp_true or expected),
                (False, exp_false or expected),
            ):
                with self.subTest(paths, use_directory_urls=use_directory_urls):
                    src_path, dest_path = paths
                    f = File(src_path, '', '', use_directory_urls)
                    actual = _RelativePathTreeprocessor._possible_target_uris(
                        f, dest_path, use_directory_urls
                    )
                    self.assertEqual(list(actual), expected_paths.split(', '))

        test(('index.md', 'index.md'), expected='index.md')
        test(('index.md', 'foo/bar.md'), expected='foo/bar.md')
        test(
            ('index.md', 'foo/bar'),
            expected='foo/bar, foo/bar/index.md, foo/bar/README.md, foo/bar.md',
        )

        test(('index.md', 'foo/bar.html'), expected='foo/bar.html, foo/bar.md')
        test(
            ('foo.md', 'foo/bar.html'),
            exp_true='foo/bar.html, foo/bar.md, foo/foo/bar.html, foo/foo/bar.md',
            exp_false='foo/bar.html, foo/bar.md',
        )

        test(('foo.md', 'index.md'), exp_true='index.md, foo/index.md', exp_false='index.md')
        test(('foo.md', 'foo.md'), exp_true='foo.md, foo/foo.md', exp_false='foo.md')
        test(('foo.md', 'bar.md'), exp_true='bar.md, foo/bar.md', exp_false='bar.md')
        test(
            ('foo.md', 'foo/bar.md'), exp_true='foo/bar.md, foo/foo/bar.md', exp_false='foo/bar.md'
        )
        test(
            ('foo.md', 'foo'),
            exp_true='foo, foo/index.md, foo/README.md, foo.md, foo/foo, foo/foo/index.md, foo/foo/README.md, foo/foo.md',
            exp_false='foo, foo/index.md, foo/README.md, foo.md',
        )
        test(
            ('foo.md', 'foo/bar'),
            exp_true='foo/bar, foo/bar/index.md, foo/bar/README.md, foo/bar.md, foo/foo/bar, foo/foo/bar/index.md, foo/foo/bar/README.md, foo/foo/bar.md',
            exp_false='foo/bar, foo/bar/index.md, foo/bar/README.md, foo/bar.md',
        )
        test(
            ('foo.md', 'foo/bar/'),
            exp_true='foo/bar, foo/bar/index.md, foo/bar/README.md, foo/bar.md, foo/foo/bar, foo/foo/bar/index.md, foo/foo/bar/README.md, foo/foo/bar.md',
            exp_false='foo/bar, foo/bar/index.md, foo/bar/README.md',
        )

        test(
            ('foo.md', 'foo.html'),
            exp_true='foo.html, foo.md, foo/foo.html, foo/foo.md',
            exp_false='foo.html, foo.md',
        )
        test(
            ('foo.md', '../foo/'),
            exp_true='../foo, foo, foo/index.md, foo/README.md, foo.md',
            exp_false='../foo',
        )
        test(
            ('foo.md', 'foo/'),
            exp_true='foo, foo/index.md, foo/README.md, foo.md, foo/foo, foo/foo/index.md, foo/foo/README.md, foo/foo.md',
            exp_false='foo, foo/index.md, foo/README.md',
        )
        test(('foo/index.md', 'index.md'), expected='foo/index.md')
        test(('foo/index.md', 'foo/bar.html'), expected='foo/foo/bar.html, foo/foo/bar.md')
        test(('foo/index.md', '../foo.html'), expected='foo.html, foo.md')
        test(('foo/index.md', '../'), expected='., index.md, README.md')
        test(
            ('foo/bar.md', 'index.md'),
            exp_true='foo/index.md, foo/bar/index.md',
            exp_false='foo/index.md',
        )
        test(
            ('foo/bar.md', 'foo.md'),
            exp_true='foo/foo.md, foo/bar/foo.md',
            exp_false='foo/foo.md',
        )
        test(
            ('foo/bar.md', 'bar.md'),
            exp_true='foo/bar.md, foo/bar/bar.md',
            exp_false='foo/bar.md',
        )
        test(
            ('foo/bar.md', 'foo/bar.md'),
            exp_true='foo/foo/bar.md, foo/bar/foo/bar.md',
            exp_false='foo/foo/bar.md',
        )
        test(
            ('foo/bar.md', 'foo'),
            exp_true='foo/foo, foo/foo/index.md, foo/foo/README.md, foo/foo.md, foo/bar/foo, foo/bar/foo/index.md, foo/bar/foo/README.md, foo/bar/foo.md',
            exp_false='foo/foo, foo/foo/index.md, foo/foo/README.md, foo/foo.md',
        )
        test(
            ('foo/bar.md', 'foo/bar'),
            exp_true='foo/foo/bar, foo/foo/bar/index.md, foo/foo/bar/README.md, foo/foo/bar.md, foo/bar/foo/bar, foo/bar/foo/bar/index.md, foo/bar/foo/bar/README.md, foo/bar/foo/bar.md',
            exp_false='foo/foo/bar, foo/foo/bar/index.md, foo/foo/bar/README.md, foo/foo/bar.md',
        )
        test(
            ('foo/bar.md', 'foo.html'),
            exp_true='foo/foo.html, foo/foo.md, foo/bar/foo.html, foo/bar/foo.md',
            exp_false='foo/foo.html, foo/foo.md',
        )
        test(
            ('foo/bar.md', 'foo/bar.html'),
            exp_true='foo/foo/bar.html, foo/foo/bar.md, foo/bar/foo/bar.html, foo/bar/foo/bar.md',
            exp_false='foo/foo/bar.html, foo/foo/bar.md',
        )
        test(
            ('foo/bar.md', '../foo/bar.html'),
            exp_true='foo/bar.html, foo/bar.md, foo/foo/bar.html, foo/foo/bar.md',
            exp_false='foo/bar.html, foo/bar.md',
        )
        test(
            ('foo/bar.md', '../foo'),
            exp_true='foo, foo/index.md, foo/README.md, foo.md, foo/foo, foo/foo/index.md, foo/foo/README.md, foo/foo.md',
            exp_false='foo, foo/index.md, foo/README.md, foo.md',
        )
        test(
            ('foo/bar.md', '../'),
            exp_true='., index.md, README.md, foo, foo/index.md, foo/README.md',
            exp_false='., index.md, README.md',
        )

        for src in 'foo/bar.md', 'foo.md', 'foo/index.md':
            test((src, '/foo'), expected='foo, foo/index.md, foo/README.md, foo.md')
            test((src, '/foo/bar.md'), expected='foo/bar.md')
            test((src, '/foo/bar.html'), expected='foo/bar.html, foo/bar.md')

        for dest in '', '.', './':
            test(('index.md', dest), expected='., index.md')
            test(('foo/bar.md', dest), expected='foo, foo/bar.md')

        test(
            ('foo/bar.md', '../test.png'),
            exp_true='test.png, test.png.md, foo/test.png, foo/test.png.md',
            exp_false='test.png, test.png.md',
        )



import datetime
import urllib.request
from typing import Callable

import mkdocs_get_deps.cache

import mkdocs


def download_url(url: str) -> bytes:
    req = urllib.request.Request(url, headers={"User-Agent": f"mkdocs/{mkdocs.__version__}"})
    with urllib.request.urlopen(req) as resp:
        return resp.read()


def download_and_cache_url(
    url: str,
    cache_duration: datetime.timedelta,
    *,
    download: Callable[[str], bytes] = download_url,
    comment: bytes = b"# ",
) -> bytes:
    """
    Downloads a file from the URL, stores it under ~/.cache/, and returns its content.

    For tracking the age of the content, a prefix is inserted into the stored file, rather than relying on mtime.

    Args:
        url: URL to use.
        download: Callback that will accept the URL and actually perform the download.
        cache_duration: How long to consider the URL content cached.
        comment: The appropriate comment prefix for this file format.
    """
    return mkdocs_get_deps.cache.download_and_cache_url(
        url=url, cache_duration=cache_duration, download=download, comment=comment
    )



#!/usr/bin/env python

from __future__ import annotations

import logging
import os
import shutil
import sys
import textwrap
import traceback
import warnings

import click

from mkdocs import __version__, config, utils

if sys.platform.startswith("win"):
    try:
        import colorama
    except ImportError:
        pass
    else:
        colorama.init()

log = logging.getLogger(__name__)


def _showwarning(message, category, filename, lineno, file=None, line=None):
    try:
        # Last stack frames:
        # * ...
        # * Location of call to deprecated function   <-- include this
        # * Location of call to warn()                <-- include this
        # * (stdlib) Location of call to showwarning function
        # * (this function) Location of call to extract_stack()
        stack = [frame for frame in traceback.extract_stack() if frame.line][-4:-2]
        # Make sure the actual affected file's name is still present (the case of syntax warning):
        if not any(frame.filename == filename for frame in stack):
            stack = stack[-1:] + [traceback.FrameSummary(filename, lineno, '')]

        tb = ''.join(traceback.format_list(stack))
    except Exception:
        tb = f'  File "{filename}", line {lineno}'

    log.info(f'{category.__name__}: {message}\n{tb}')


def _enable_warnings():
    from mkdocs.commands import build

    build.log.addFilter(utils.DuplicateFilter())

    warnings.simplefilter('module', DeprecationWarning)
    warnings.showwarning = _showwarning


class ColorFormatter(logging.Formatter):
    colors = {
        'CRITICAL': 'red',
        'ERROR': 'red',
        'WARNING': 'yellow',
        'DEBUG': 'blue',
    }

    text_wrapper = textwrap.TextWrapper(
        width=shutil.get_terminal_size(fallback=(0, 0)).columns,
        replace_whitespace=False,
        break_long_words=False,
        break_on_hyphens=False,
        initial_indent=' ' * 11,
        subsequent_indent=' ' * 11,
    )

    def format(self, record):
        message = super().format(record)
        prefix = f'{record.levelname:<8}-  '
        if record.levelname in self.colors:
            prefix = click.style(prefix, fg=self.colors[record.levelname])
        if self.text_wrapper.width:
            # Only wrap text if a terminal width was detected
            msg = '\n'.join(self.text_wrapper.fill(line) for line in message.splitlines())
            # Prepend prefix after wrapping so that color codes don't affect length
            return prefix + msg[11:]
        return prefix + message


class State:
    """Maintain logging level."""

    def __init__(self, log_name='mkdocs', level=logging.INFO):
        self.logger = logging.getLogger(log_name)
        self.logger.setLevel(level)
        self.logger.propagate = False

        self.stream = logging.StreamHandler()
        self.stream.setFormatter(ColorFormatter())
        self.stream.name = 'MkDocsStreamHandler'
        self.logger.addHandler(self.stream)

    def __del__(self):
        self.logger.removeHandler(self.stream)


pass_state = click.make_pass_decorator(State, ensure=True)

clean_help = "Remove old files from the site_dir before building (the default)."
config_help = (
    "Provide a specific MkDocs config. This can be a file name, or '-' to read from stdin."
)
dev_addr_help = "IP address and port to serve documentation locally (default: localhost:8000)"
serve_open_help = "Open the website in a Web browser after the initial build finishes."
strict_help = "Enable strict mode. This will cause MkDocs to abort the build on any warnings."
theme_help = "The theme to use when building your documentation."
theme_choices = sorted(utils.get_theme_names())
site_dir_help = "The directory to output the result of the documentation build."
use_directory_urls_help = "Use directory URLs when building pages (the default)."
reload_help = "Enable the live reloading in the development server (this is the default)"
no_reload_help = "Disable the live reloading in the development server."
serve_dirty_help = "Only re-build files that have changed."
serve_clean_help = (
    "Build the site without any effects of `mkdocs serve` - pure `mkdocs build`, then serve."
)
commit_message_help = (
    "A commit message to use when committing to the "
    "GitHub Pages remote branch. Commit {sha} and MkDocs {version} are available as expansions"
)
remote_branch_help = (
    "The remote branch to commit to for GitHub Pages. This "
    "overrides the value specified in config"
)
remote_name_help = (
    "The remote name to commit to for GitHub Pages. This overrides the value specified in config"
)
force_help = "Force the push to the repository."
no_history_help = "Replace the whole Git history with one new commit."
ignore_version_help = (
    "Ignore check that build is not being deployed with an older version of MkDocs."
)
watch_theme_help = (
    "Include the theme in list of files to watch for live reloading. "
    "Ignored when live reload is not used."
)
shell_help = "Use the shell when invoking Git."
watch_help = "A directory or file to watch for live reloading. Can be supplied multiple times."
projects_file_help = (
    "URL or local path of the registry file that declares all known MkDocs-related projects."
)


def add_options(*opts):
    def inner(f):
        for i in reversed(opts):
            f = i(f)
        return f

    return inner


def verbose_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        if value:
            state.logger.setLevel(logging.DEBUG)

    return click.option(
        '-v',
        '--verbose',
        is_flag=True,
        expose_value=False,
        help='Enable verbose output',
        callback=callback,
    )(f)


def quiet_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        if value:
            state.logger.setLevel(logging.ERROR)

    return click.option(
        '-q',
        '--quiet',
        is_flag=True,
        expose_value=False,
        help='Silence warnings',
        callback=callback,
    )(f)


def color_option(f):
    def callback(ctx, param, value):
        state = ctx.ensure_object(State)
        if value is False or (
            value is None
            and (
                not sys.stdout.isatty()
                or os.environ.get('NO_COLOR')
                or os.environ.get('TERM') == 'dumb'
            )
        ):
            state.stream.setFormatter(logging.Formatter('%(levelname)-8s-  %(message)s'))

    return click.option(
        '--color/--no-color',
        is_flag=True,
        default=None,
        expose_value=False,
        help="Force enable or disable color and wrapping for the output. Default is auto-detect.",
        callback=callback,
    )(f)


common_options = add_options(quiet_option, verbose_option)
common_config_options = add_options(
    click.option('-f', '--config-file', type=click.File('rb'), help=config_help),
    # Don't override config value if user did not specify --strict flag
    # Conveniently, load_config drops None values
    click.option('-s', '--strict/--no-strict', is_flag=True, default=None, help=strict_help),
    click.option('-t', '--theme', type=click.Choice(theme_choices), help=theme_help),
    # As with --strict, set the default to None so that this doesn't incorrectly
    # override the config file
    click.option(
        '--use-directory-urls/--no-directory-urls',
        is_flag=True,
        default=None,
        help=use_directory_urls_help,
    ),
)

PYTHON_VERSION = f"{sys.version_info.major}.{sys.version_info.minor}"

PKG_DIR = os.path.dirname(os.path.abspath(__file__))


@click.group(context_settings=dict(help_option_names=['-h', '--help'], max_content_width=120))
@click.version_option(
    __version__,
    '-V',
    '--version',
    message=f'%(prog)s, version %(version)s from { PKG_DIR } (Python { PYTHON_VERSION })',
)
@common_options
@color_option
def cli():
    """MkDocs - Project documentation with Markdown."""


@cli.command(name="serve")
@click.option('-a', '--dev-addr', help=dev_addr_help, metavar='<IP:PORT>')
@click.option('-o', '--open', 'open_in_browser', help=serve_open_help, is_flag=True)
@click.option('--no-livereload', 'livereload', flag_value=False, help=no_reload_help)
@click.option('--livereload', 'livereload', flag_value=True, default=True, hidden=True)
@click.option('--dirtyreload', 'build_type', flag_value='dirty', hidden=True)
@click.option('--dirty', 'build_type', flag_value='dirty', help=serve_dirty_help)
@click.option('-c', '--clean', 'build_type', flag_value='clean', help=serve_clean_help)
@click.option('--watch-theme', help=watch_theme_help, is_flag=True)
@click.option(
    '-w', '--watch', help=watch_help, type=click.Path(exists=True), multiple=True, default=[]
)
@common_config_options
@common_options
def serve_command(**kwargs):
    """Run the builtin development server."""
    from mkdocs.commands import serve

    _enable_warnings()
    serve.serve(**kwargs)


@cli.command(name="build")
@click.option('-c', '--clean/--dirty', is_flag=True, default=True, help=clean_help)
@common_config_options
@click.option('-d', '--site-dir', type=click.Path(), help=site_dir_help)
@common_options
def build_command(clean, **kwargs):
    """Build the MkDocs documentation."""
    from mkdocs.commands import build

    _enable_warnings()
    cfg = config.load_config(**kwargs)
    cfg.plugins.on_startup(command='build', dirty=not clean)
    try:
        build.build(cfg, dirty=not clean)
    finally:
        cfg.plugins.on_shutdown()


@cli.command(name="gh-deploy")
@click.option('-c', '--clean/--dirty', is_flag=True, default=True, help=clean_help)
@click.option('-m', '--message', help=commit_message_help)
@click.option('-b', '--remote-branch', help=remote_branch_help)
@click.option('-r', '--remote-name', help=remote_name_help)
@click.option('--force', is_flag=True, help=force_help)
@click.option('--no-history', is_flag=True, help=no_history_help)
@click.option('--ignore-version', is_flag=True, help=ignore_version_help)
@click.option('--shell', is_flag=True, help=shell_help)
@common_config_options
@click.option('-d', '--site-dir', type=click.Path(), help=site_dir_help)
@common_options
def gh_deploy_command(
    clean, message, remote_branch, remote_name, force, no_history, ignore_version, shell, **kwargs
):
    """Deploy your documentation to GitHub Pages."""
    from mkdocs.commands import build, gh_deploy

    _enable_warnings()
    cfg = config.load_config(remote_branch=remote_branch, remote_name=remote_name, **kwargs)
    cfg.plugins.on_startup(command='gh-deploy', dirty=not clean)
    try:
        build.build(cfg, dirty=not clean)
    finally:
        cfg.plugins.on_shutdown()
    gh_deploy.gh_deploy(
        cfg,
        message=message,
        force=force,
        no_history=no_history,
        ignore_version=ignore_version,
        shell=shell,
    )


@cli.command(name="get-deps")
@verbose_option
@click.option('-f', '--config-file', type=click.File('rb'), help=config_help)
@click.option(
    '-p',
    '--projects-file',
    default=None,
    help=projects_file_help,
    show_default=True,
)
def get_deps_command(config_file, projects_file):
    """Show required PyPI packages inferred from plugins in mkdocs.yml."""
    from mkdocs_get_deps import get_deps, get_projects_file

    from mkdocs.config.base import _open_config_file

    warning_counter = utils.CountHandler()
    warning_counter.setLevel(logging.WARNING)
    logging.getLogger('mkdocs').addHandler(warning_counter)

    with get_projects_file(projects_file) as p:
        with _open_config_file(config_file) as f:
            deps = get_deps(config_file=f, projects_file=p)

    for dep in deps:
        print(dep)  # noqa: T201

    if warning_counter.get_counts():
        sys.exit(1)


@cli.command(name="new")
@click.argument("project_directory")
@common_options
def new_command(project_directory):
    """Create a new MkDocs project."""
    from mkdocs.commands import new

    new.new(project_directory)


if __name__ == '__main__':  # pragma: no cover
    cli()



from __future__ import annotations

import logging
from typing import IO, Dict, Mapping

from mkdocs.config import base
from mkdocs.config import config_options as c
from mkdocs.structure.pages import Page, _AbsoluteLinksValidationValue
from mkdocs.utils.yaml import get_yaml_loader, yaml_load


class _LogLevel(c.OptionallyRequired[int]):
    levels: Mapping[str, int] = {
        "warn": logging.WARNING,
        "info": logging.INFO,
        "ignore": logging.DEBUG,
    }

    def run_validation(self, value: object) -> int:
        if not isinstance(value, str):
            raise base.ValidationError(f"Expected a string, but a {type(value)} was given.")
        try:
            return self.levels[value]
        except KeyError:
            raise base.ValidationError(f"Expected one of {list(self.levels)}, got {value!r}")


class _AbsoluteLinksValidation(_LogLevel):
    levels: Mapping[str, int] = {
        **_LogLevel.levels,
        "relative_to_docs": _AbsoluteLinksValidationValue.RELATIVE_TO_DOCS,
    }


# NOTE: The order here is important. During validation some config options
# depend on others. So, if config option A depends on B, then A should be
# listed higher in the schema.
class MkDocsConfig(base.Config):
    """The configuration of MkDocs itself (the root object of mkdocs.yml)."""

    config_file_path: str = c.Type(str)  # type: ignore[assignment]
    """The path to the mkdocs.yml config file. Can't be populated from the config."""

    site_name = c.Type(str)
    """The title to use for the documentation."""

    nav = c.Optional(c.Nav())
    """Defines the structure of the navigation."""
    pages = c.Deprecated(removed=True, moved_to='nav')

    exclude_docs = c.Optional(c.PathSpec())
    """Gitignore-like patterns of files (relative to docs dir) to exclude from the site."""

    draft_docs = c.Optional(c.PathSpec())
    """Gitignore-like patterns of files (relative to docs dir) to mark as draft."""

    not_in_nav = c.Optional(c.PathSpec())
    """Gitignore-like patterns of files (relative to docs dir) that are not intended to be in the nav.

    This marks doc files that are expected not to be in the nav, otherwise they will cause a log message
    (see also `validation.nav.omitted_files`).
    """

    site_url = c.Optional(c.URL(is_dir=True))
    """The full URL to where the documentation will be hosted."""

    site_description = c.Optional(c.Type(str))
    """A description for the documentation project that will be added to the
    HTML meta tags."""
    site_author = c.Optional(c.Type(str))
    """The name of the author to add to the HTML meta tags."""

    theme = c.Theme(default='mkdocs')
    """The MkDocs theme for the documentation."""

    docs_dir = c.DocsDir(default='docs', exists=True)
    """The directory containing the documentation markdown."""

    site_dir = c.SiteDir(default='site')
    """The directory where the site will be built to"""

    copyright = c.Optional(c.Type(str))
    """A copyright notice to add to the footer of documentation."""

    google_analytics = c.Deprecated(
        message=(
            'The configuration option {} has been deprecated and '
            'will be removed in a future release of MkDocs. See the '
            'options available on your theme for an alternative.'
        ),
        option_type=c.Type(list, length=2),
    )
    """set of values for Google analytics containing the account IO and domain
    this should look like, ['UA-27795084-5', 'mkdocs.org']"""

    dev_addr = c.IpAddress(default='127.0.0.1:8000')
    """The address on which to serve the live reloading docs server."""

    use_directory_urls = c.Type(bool, default=True)
    """If `True`, use `<page_name>/index.html` style files with hyperlinks to
    the directory. If `False`, use `<page_name>.html style file with
    hyperlinks to the file.
    True generates nicer URLs, but False is useful if browsing the output on
    a filesystem."""

    repo_url = c.Optional(c.URL())
    """Specify a link to the project source repo to be included
    in the documentation pages."""

    repo_name = c.Optional(c.RepoName('repo_url'))
    """A name to use for the link to the project source repo.
    Default, If repo_url is unset then None, otherwise
    "GitHub", "Bitbucket" or "GitLab" for known url or Hostname
    for unknown urls."""

    edit_uri_template = c.Optional(c.EditURITemplate('edit_uri'))
    edit_uri = c.Optional(c.EditURI('repo_url'))
    """Specify a URI to the docs dir in the project source repo, relative to the
    repo_url. When set, a link directly to the page in the source repo will
    be added to the generated HTML. If repo_url is not set also, this option
    is ignored."""

    extra_css = c.Type(list, default=[])
    extra_javascript = c.ListOfItems(c.ExtraScript(), default=[])
    """Specify which css or javascript files from the docs directory should be
    additionally included in the site."""

    extra_templates = c.Type(list, default=[])
    """Similar to the above, but each template (HTML or XML) will be build with
    Jinja2 and the global context."""

    markdown_extensions = c.MarkdownExtensions(
        builtins=['toc', 'tables', 'fenced_code'], configkey='mdx_configs'
    )
    """PyMarkdown extension names."""

    mdx_configs = c.Private[Dict[str, dict]]()
    """PyMarkdown extension configs. Populated from `markdown_extensions`."""

    strict = c.Type(bool, default=False)
    """Enabling strict mode causes MkDocs to stop the build when a problem is
    encountered rather than display an error."""

    remote_branch = c.Type(str, default='gh-pages')
    """The remote branch to commit to when using gh-deploy."""

    remote_name = c.Type(str, default='origin')
    """The remote name to push to when using gh-deploy."""

    extra = c.SubConfig()
    """extra is a mapping/dictionary of data that is passed to the template.
    This allows template authors to require extra configuration that not
    relevant to all themes and doesn't need to be explicitly supported by
    MkDocs itself. A good example here would be including the current
    project version."""

    plugins = c.Plugins(theme_key='theme', default=['search'])
    """A list of plugins. Each item may contain a string name or a key value pair.
    A key value pair should be the string name (as the key) and a dict of config
    options (as the value)."""

    hooks = c.Hooks('plugins')
    """A list of filenames that will be imported as Python modules and used as
    an instance of a plugin each."""

    watch = c.ListOfPaths(default=[])
    """A list of extra paths to watch while running `mkdocs serve`."""

    class Validation(base.Config):
        class NavValidation(base.Config):
            omitted_files = _LogLevel(default='info')
            """Warning level for when a doc file is never mentioned in the navigation.
            For granular configuration, see `not_in_nav`."""

            not_found = _LogLevel(default='warn')
            """Warning level for when the navigation links to a relative path that isn't an existing page on the site."""

            absolute_links = _AbsoluteLinksValidation(default='info')
            """Warning level for when the navigation links to an absolute path (starting with `/`)."""

        nav = c.SubConfig(NavValidation)

        class LinksValidation(base.Config):
            not_found = _LogLevel(default='warn')
            """Warning level for when a Markdown doc links to a relative path that isn't an existing document on the site."""

            absolute_links = _AbsoluteLinksValidation(default='info')
            """Warning level for when a Markdown doc links to an absolute path (starting with `/`)."""

            unrecognized_links = _LogLevel(default='info')
            """Warning level for when a Markdown doc links to a relative path that doesn't look like
            it could be a valid internal link. For example, if the link ends with `/`."""

            anchors = _LogLevel(default='info')
            """Warning level for when a Markdown doc links to an anchor that's not present on the target page."""

        links = c.SubConfig(LinksValidation)

    validation = c.PropagatingSubConfig[Validation]()

    _current_page: Page | None = None
    """The currently rendered page. Please do not access this and instead
    rely on the `page` argument to event handlers."""

    def load_dict(self, patch: dict) -> None:
        super().load_dict(patch)
        if 'config_file_path' in patch:
            raise base.ValidationError("Can't set config_file_path in config")

    def load_file(self, config_file: IO) -> None:
        """Load config options from the open file descriptor of a YAML file."""
        loader = get_yaml_loader(config=self)
        self.load_dict(yaml_load(config_file, loader))


def get_schema() -> base.PlainConfigSchema:
    """Soft-deprecated, do not use."""
    return MkDocsConfig._schema
